import asyncio
import json
import re
from typing import Any, Generator, Optional, Type, Union, cast

import mercantile
import numpy as np
from httpx import Response
from shapely import MultiPolygon, Point, Polygon, to_geojson

from pynspd.client import get_async_client
from pynspd.errors import TooBigContour
from pynspd.schemas import Layer36048Feature, Layer36049Feature, NspdFeature
from pynspd.schemas.feature import Feat
from pynspd.schemas.responses import SearchResponse
from pynspd.types.enums import ThemeId


class AsyncNspd:
    def __init__(self):
        self._client = get_async_client()

    async def __aenter__(self):
        return self

    async def __aexit__(self, *exc):
        await self.close()

    async def close(self):
        """Окончание сессии"""
        await self._client.aclose()

    @staticmethod
    def iter_cn(input_str: str) -> Generator[str, None, None]:
        """Извлечение кадастровых номеров из строки"""
        for cn in re.findall(r"\d+:\d+:\d+:\d+", input_str):
            yield cn

    @staticmethod
    def _cast_feature_to_layer_def(
        raw_feature: Optional[NspdFeature], layer_def: Type[Feat]
    ) -> Optional[Feat]:
        """Приведение базовой фичи к фиче конкретного слоя"""
        if raw_feature is None:
            return None
        feature = raw_feature.cast(layer_def)
        return feature

    @staticmethod
    def _cast_features_to_layer_defs(
        raw_features: Optional[list[NspdFeature]], layer_def: Type[Feat]
    ) -> Optional[list[Feat]]:
        """Аналог `_cast_feature_to_layer_def` для списка фичей"""
        if raw_features is None:
            return None
        features = [i.cast(layer_def) for i in raw_features]
        return features

    @staticmethod
    def _validate_response(response: Response) -> Optional[list[NspdFeature]]:
        response.raise_for_status()
        features = response.json()["features"]
        if len(features) == 0:
            return None
        return [NspdFeature.model_validate(i) for i in features]

    async def _search(self, params: dict[str, Any]) -> Optional[SearchResponse]:
        r = await self._client.get("/api/geoportal/v2/search/geoportal", params=params)
        if r.status_code == 404:
            return None
        r.raise_for_status()
        return SearchResponse.model_validate(r.json())

    async def _search_one(self, params: dict[str, Any]) -> Optional[NspdFeature]:
        response = await self._search(params)
        if response is None:
            return None
        features = response.data.features
        assert len(features) == 1
        return features[0]

    async def search_by_theme(
        self, query: str, theme_id: ThemeId = ThemeId.REAL_ESTATE_OBJECTS
    ) -> Optional[NspdFeature]:
        """Глобальный поисковой запрос

        Args:
            query (str): поисковой запрос
            theme_id (int): вид объекта (кадастровое деление, объект недвижимости и т.д.)

        Returns:
            Optional[SearchResponse]: положительный ответ от сервиса, либо None, если ничего не найдено
        """
        return await self._search_one(
            params={
                "query": query,
                "thematicSearchId": theme_id.value,
            }
        )

    async def search_by_layers(
        self, query: str, *layer_ids: int
    ) -> Optional[NspdFeature]:
        """Поисковой запрос по указанным слоям

        Args:
            query (str): поисковой запрос
            *layer_ids (int): id слоев, в которых будет производиться поиск

        Returns:
            Optional[SearchResponse]: положительный ответ от сервиса, либо None, если ничего не найдено
        """
        return await self._search_one(
            params={
                "query": query,
                "layersId": layer_ids,
            }
        )

    async def search_by_model(
        self, query: str, layer_def: Type[Feat]
    ) -> Optional[Feat]:
        """Поиск одного объекта по определению слоя

        Определение слоя можно
        1) импортировать из `pynspd.schemas`, зная его id
        ```
        from pynspd.schemas import Layer36048Feature    # 36048 - id слоя, которое мы подсмотрели из запросов к wms на сайте
        feature = await api.search_by_model("77:06:0004002:7207", Layer36048Feature)
        feature.properties.options.cad_num    # IDE знает тип и подсказывает возможные свойства
        ```

        2) воспользоваться методом `NspdFeature.by_title`
        ```
        from pynspd.schemas import NspdFeature
        feature = await api.search_by_model("77:06:0004002:7207", NspdFeature.by_title("Земельные участки из ЕГРН"))    # IDE знает весь перечень слоев и подсказывает ввод
        feature.properties.options.cad_num    # свойство будет так же доступно, но IDE уже не знает о нем
        ```

        Args:
            query (str): поисковой запрос
            layer_def (Type[Feat]): Определение слоя

        Returns:
            Optional[Feat]: валидированная модель слоя, если найдено
        """
        feature = await self.search_by_layers(query, layer_def.layer_meta.layer_id)
        return self._cast_feature_to_layer_def(feature, layer_def)

    async def search_zu(self, cn: str) -> Optional[Layer36048Feature]:
        """Поиск ЗУ по кадастровому номеру"""
        layer_def = cast(
            Type[Layer36048Feature], NspdFeature.by_title("Земельные участки из ЕГРН")
        )
        return await self.search_by_model(cn, layer_def)

    async def search_many_zu(
        self, cns_string: str
    ) -> list[Optional[Layer36048Feature]]:
        """Поиск всех ЗУ, содержащихся в строке"""
        cns = list(self.iter_cn(cns_string))
        features = await asyncio.gather(*[self.search_zu(cn) for cn in cns])
        return features

    async def search_oks(self, cn: str) -> Optional[Layer36049Feature]:
        """Поиск ОКС по кадастровому номеру"""
        layer_def = cast(Type[Layer36049Feature], NspdFeature.by_title("Здания"))
        return await self.search_by_model(cn, layer_def)

    async def search_many_oks(
        self, cns_string: str
    ) -> list[Optional[Layer36049Feature]]:
        """Поиск всех ОКС, содержащихся в строке"""
        cns = list(self.iter_cn(cns_string))
        features = await asyncio.gather(*[self.search_oks(cn) for cn in cns])
        return features

    async def search_in_contour(
        self,
        countour: Union[Polygon, MultiPolygon],
        *category_ids: int,
        epsg: int = 4326,
    ) -> Optional[list[NspdFeature]]:
        """Поиск объектов в контуре по id категорий слоев

        Args:
            countour (Union[Polygon, MultiPolygon]): Геометрический объект с контуром
            category_ids (int): id категорий слоев
            epsg (int, optional): Система координат контура. Defaults to 4326.

        Returns:
            Optional[list[Feat]]: Список объектов, пересекающихся с контуром, если найден хоть один
        """
        feature_geojson = json.loads(to_geojson(countour))
        feature_geojson["crs"] = {
            "type": "name",
            "properties": {"name": f"EPSG:{epsg}"},
        }
        payload = {
            "categories": [{"id": id_} for id_ in category_ids],
            "geom": {
                "type": "FeatureCollection",
                "features": [
                    {"geometry": feature_geojson, "type": "Feature", "properties": {}}
                ],
            },
        }
        response = await self._client.post(
            "/api/geoportal/v1/intersects",
            params={"typeIntersect": "fullObject"},
            json=payload,
        )
        if response.status_code == 500 and response.json()["code"] == 400004:
            raise TooBigContour
        return self._validate_response(response)

    async def search_in_contour_by_model(
        self,
        countour: Union[Polygon, MultiPolygon],
        layer_def: Type[Feat],
        epsg: int = 4326,
    ) -> Optional[list[Feat]]:
        """Поиск объектов в контуре по определению слоя

        Args:
            countour (Union[Polygon, MultiPolygon]): Геометрический объект с контуром
            layer_def (Type[Feat]): Модель слоя
            epsg (int, optional): Система координат контура. Defaults to 4326.

        Returns:
            Optional[list[Feat]]: Список объектов, пересекающихся с контуром, если найден хоть один
        """
        raw_features = await self.search_in_contour(
            countour, layer_def.layer_meta.category_id, epsg=epsg
        )
        return self._cast_features_to_layer_defs(raw_features, layer_def)

    async def search_zu_in_contour(
        self, countour: Union[Polygon, MultiPolygon], epsg: int = 4326
    ) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ в контуре

        Args:
            countour (Union[Polygon, MultiPolygon]): Геометрический объект с контуром
            epsg (int, optional): Система координат контура. Defaults to 4326.

        Returns:
            Optional[list[Layer36048Feature]]: Список объектов, пересекающихся с контуром, если найден хоть один
        """
        return await self.search_in_contour_by_model(
            countour, Layer36048Feature, epsg=epsg
        )

    async def search_oks_in_contour(
        self, countour: Union[Polygon, MultiPolygon], epsg: int = 4326
    ) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС в контуре

        Args:
            countour (Union[Polygon, MultiPolygon]): Геометрический объект с контуром
            epsg (int, optional): Система координат контура. Defaults to 4326.

        Returns:
            Optional[list[Layer36048Feature]]: Список объектов, пересекающихся с контуром, если найден хоть один
        """
        return await self.search_in_contour_by_model(
            countour, Layer36049Feature, epsg=epsg
        )

    async def search_at_point(
        self, pt: Point, layer_id: int
    ) -> Optional[list[NspdFeature]]:
        """Поиск объектов слоя в точке

        Args:
            pt (Point):
            layer_id (int):

        Returns:
            Optional[list[NspdFeature]]: Список объектов, если найдены
        """
        TILE_SIZE = 512
        tile = mercantile.tile(
            pt.x, pt.y, zoom=24
        )  # zoom=24 должно быть достаточно для самого точного совпадения
        tile_bounds = mercantile.bounds(tile)
        i = np.interp(pt.x, [tile_bounds.west, tile_bounds.east], [0, TILE_SIZE])
        j = np.interp(pt.y, [tile_bounds.south, tile_bounds.north], [0, TILE_SIZE])
        bbox = ",".join(
            map(str, mercantile.xy_bounds(tile))
        )  # bbox в 3857, см. комментарий про CRS
        params = {
            "REQUEST": "GetFeatureInfo",
            "SERVICE": "WMS",
            "VERSION": "1.3.0",
            "INFO_FORMAT": "application/json",
            "FORMAT": "image/png",
            "STYLES": "",
            "TRANSPARENT": "true",
            "QUERY_LAYERS": layer_id,
            "LAYERS": layer_id,
            "WIDTH": TILE_SIZE,
            "HEIGHT": TILE_SIZE,
            "I": int(i),
            "J": TILE_SIZE - int(j),  # отсчет координат для пикселей ведется сверху
            "CRS": "EPSG:3857",  # CRS для bbox
            # можно указать и 4326, но тогда и геометрия будет в 4326
            # Но в других методах мы всегда ждем 3857, поэтому оставляем
            "BBOX": bbox,
            "FEATURE_COUNT": "10",  # Если не указать - вернет только один, даже если попало на границу
        }
        response = await self._client.get(
            f"/api/aeggis/v3/{layer_id}/wms", params=params
        )
        return self._validate_response(response)

    async def search_at_point_by_model(
        self, pt: Point, layer_def: Type[Feat]
    ) -> Optional[list[Feat]]:
        """Поиск объектов слоя в точке (с типизацией)

        Args:
            pt (Point):
            layer_def (Type[Feat]): Тип слоя

        Returns:
            Optional[list[Feat]]: Типизированный список объектов, если найдены
        """
        raw_features = await self.search_at_point(pt, layer_def.layer_meta.layer_id)
        return self._cast_features_to_layer_defs(raw_features, layer_def)

    async def search_zu_at_point(self, pt: Point) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ в точке"""
        return await self.search_at_point_by_model(pt, Layer36048Feature)

    async def search_oks_at_point(self, pt: Point) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС в точке"""
        return await self.search_at_point_by_model(pt, Layer36049Feature)



import asyncio

import pytest
import pytest_asyncio
from shapely import wkt
from shapely.geometry import MultiPolygon, Point, Polygon

from pynspd import AsyncNspd, NspdFeature
from pynspd.errors import TooBigContour
from pynspd.schemas import Layer36048Feature, Layer36049Feature, Layer37578Feature


@pytest.fixture(scope="module")
def event_loop():
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()


@pytest_asyncio.fixture(scope="session")
async def api():
    async with AsyncNspd() as client:
        yield client


@pytest.mark.asyncio
async def test_search_by_theme(api: AsyncNspd):
    feat = await api.search_by_theme("77:02:0021001:5304")
    assert feat.properties.options.type == "Машино-место"


@pytest.mark.asyncio
async def test_search_by_model(api: AsyncNspd):
    lf_def = NspdFeature.by_title("ЗОУИТ объектов энергетики, связи, транспорта")
    assert lf_def == Layer37578Feature
    feat = await api.search_by_model("Останкинская телебашня", lf_def)
    assert isinstance(feat, lf_def)
    assert len(feat.properties.options.model_dump_human_readable()) > 0


@pytest.mark.asyncio
async def test_search_zu(api: AsyncNspd):
    feat = await api.search_zu("77:05:0001005:19")
    assert feat.properties.options.land_record_type == "Земельный участок"
    assert isinstance(feat.geometry.to_shape(), Polygon)
    assert isinstance(feat.geometry.to_multi_shape(), MultiPolygon)


@pytest.mark.asyncio
async def test_search_many_zu(api: AsyncNspd):
    features = await api.search_many_zu(
        "77:03:0001001:82 77:03:0001001:26 77:03:0001001:132"
    )
    assert len(features) == 3
    assert all([i is not None for i in features])


@pytest.mark.asyncio
async def test_search_oks(api: AsyncNspd):
    feat = await api.search_oks("77:03:0001001:3030")
    assert feat.properties.options.build_record_type_value == "Здание"
    assert isinstance(feat.geometry.to_shape(), Polygon)


@pytest.mark.asyncio
async def test_search_many_oks(api: AsyncNspd):
    features = await api.search_many_oks(
        "77:03:0001001:3030 77:03:0001001:1111 77:03:0001001:1112"
    )
    assert len(features) == 3
    assert all([i is not None for i in features])


@pytest.mark.asyncio
async def test_search_zu_in_contour(api: AsyncNspd):
    contour = wkt.loads(
        "Polygon ((37.62381 55.75345, 37.62577 55.75390, 37.62448 55.75278, 37.62381 55.75345))"
    )
    features = await api.search_zu_in_contour(contour)
    assert features is not None
    assert all([isinstance(i, Layer36048Feature) for i in features])
    cns = [i.properties.options.cad_num for i in features]
    assert set(["77:01:0001011:8", "77:01:0001011:14", "77:01:0001011:16"]) == set(cns)


@pytest.mark.asyncio
async def test_search_oks_in_contour(api: AsyncNspd):
    contour = wkt.loads(
        "Polygon ((37.62381 55.75345, 37.62577 55.75390, 37.62448 55.75278, 37.62381 55.75345))"
    )
    features = await api.search_oks_in_contour(contour)
    assert features is not None
    assert all([isinstance(i, Layer36049Feature) for i in features])
    cns = [i.properties.options.cad_num for i in features]
    assert set(["77:01:0001011:1164", "77:01:0001011:1002"]) == set(cns)


@pytest.mark.asyncio
async def test_search_too_big_contour(api: AsyncNspd):
    with pytest.raises(TooBigContour):
        contour = wkt.loads(
            "Polygon ((37.5658 55.8198, 37.5267 55.7710, 37.6214 55.8033, 37.5658 55.8198))"
        )
        await api.search_zu_in_contour(contour)


@pytest.mark.asyncio
async def test_search_in_contour_empty(api: AsyncNspd):
    contour = wkt.loads(
        "Polygon ((37.63215 55.75588, 37.63214 55.75557, 37.63271 55.75570, 37.63215 55.75588))"
    )
    features = await api.search_zu_in_contour(contour)
    assert features is None


@pytest.mark.asyncio
async def test_search_zu_at_point(api: AsyncNspd):
    features = await api.search_zu_at_point(Point(37.54658156, 55.78729561))
    assert features is None
    features = await api.search_zu_at_point(Point(37.546440653, 55.787139958))
    assert features is not None and len(features) == 1
    assert features[0].properties.options.cad_num == "77:09:0005008:11446"


@pytest.mark.asyncio
async def test_search_oks_at_point(api: AsyncNspd):
    features = await api.search_oks_at_point(Point(37.547790951, 55.786436698))
    assert features is None
    features = await api.search_oks_at_point(Point(37.547785813, 55.786436698))
    assert features is not None and len(features) == 1
    assert features[0].properties.options.cad_num == "77:09:0005014:1044"



from pynspd._async.api import AsyncNspd
from pynspd._sync.api import Nspd
from pynspd.errors import UnknownLayer
from pynspd.schemas import NspdFeature
from pynspd.types.enums import ThemeId

__all__ = ["AsyncNspd", "Nspd", "NspdFeature", "ThemeId", "UnknownLayer"]



import pytest

from pynspd import Nspd, UnknownLayer
from pynspd.schemas import Layer36049Feature, Options36383
from pynspd.schemas.feature import _BaseFeature


def find_and_cast(cn: str):
    with Nspd() as api:
        feat = api.search_by_theme(cn)
        assert feat is not None
        cf = feat.cast()
        return cf


def test_feature_cast():
    cf = find_and_cast("77:01:0001044:1033")
    assert isinstance(cf, Layer36049Feature)
    assert len(cf.properties.options.model_dump_human_readable()) > 0


def test_feature_cast_by_category():
    cf = find_and_cast("77:01:0001044:2938")
    assert isinstance(cf, _BaseFeature)
    assert isinstance(cf.properties.options, Options36383)
    assert len(cf.properties.options.model_dump_human_readable()) > 0


def test_feature_failed_cast_on_hidden_layer():
    with pytest.raises(UnknownLayer):
        find_and_cast("77:02:0021001:5304")



from datetime import datetime
from typing import Annotated, Generic, Optional, TypeVar

from pydantic import BaseModel, ConfigDict, Field

from pynspd.schemas._common import CamelModel

OptProps = TypeVar("OptProps", bound="OptionProperties")


class Properties(CamelModel, Generic[OptProps]):
    category: int
    category_name: str
    options: OptProps
    systemInfo: Optional["SystemInfoProperties"] = None
    cadastral_districts_code: Optional[int] = None
    descr: Optional[str] = None
    external_key: Optional[str] = None
    interaction_id: Optional[int] = None
    label: Optional[str] = None
    subcategory: Optional[int] = None


class OptionProperties(BaseModel):
    model_config = ConfigDict(extra="allow", use_attribute_docstrings=True)

    no_coords: Annotated[bool, Field(alias="geocoderObject", default=False)]

    def model_dump_human_readable(self):
        """Генерация словаря с ключами, аналогичным карточке на сайте"""
        data = self.model_dump()
        alias = {
            k: v.description for k, v in self.model_fields.items() if v.description
        }
        aliased_data = {alias[k]: v for k, v in data.items() if k in alias}
        return aliased_data


class SystemInfoProperties(CamelModel):
    inserted: datetime
    inserted_by: str
    updated: datetime
    updated_by: str



from typing import Union

from geojson_pydantic import FeatureCollection
from pydantic import RootModel

from pynspd.schemas import NspdFeature
from pynspd.schemas._common import CamelModel


class BadResponse(CamelModel):
    code: int
    message: str
    request_id: str


class SearchResponse(CamelModel):
    data: FeatureCollection[NspdFeature]
    meta: list["Meta"]


class Meta(CamelModel):
    total_count: int
    category_id: int


ResponseModel = RootModel[Union[SearchResponse, BadResponse]]



class UnknownBadSearchResponse(Exception):
    """Неожиданный ответ при поиске"""


class UnknownLayer(Exception):
    """Не найдено определение слоя (по имени, категории и тд)"""


class TooBigContour(Exception):
    """Контур слишком большой, сервер не может ответить"""

    def __init__(self):
        super().__init__("Попробуйте уменьшить площадь поиска")



from typing import Any, Literal, Optional

from pydantic import ConfigDict, field_serializer, field_validator

from pynspd.schemas._common import CamelModel


class LayersTree(CamelModel):
    layers: list["LayerNode"]
    tree: Any


class LayerNode(CamelModel):
    title: str
    layer_tree_id: int
    layer_id: int
    layer_type: Literal["wms"]
    geometry_type: Literal["Polygon", "LineString", "Point", "GeometryCollection"]
    layer_name: str
    layer_visible_by_default: bool
    category_id: int

    @field_validator("geometry_type", mode="before")
    @classmethod
    def _geometry_type_validator(cls, v: str):
        alias_dict = {
            "POLYGON": "Polygon",
            "LINESTRING": "LineString",
            "POINT": "Point",
            "MULTITYPE": "GeometryCollection",
        }
        return alias_dict[v]

    @field_serializer("geometry_type")
    @classmethod
    def _geometry_type_serializer(cls, v: str):
        alias_dict = {
            "Polygon": "POLYGON",
            "LineString": "LINESTRING",
            "Point": "POINT",
            "GeometryCollection": "MULTITYPE",
        }
        return alias_dict[v]


class Card(CamelModel):
    title: Any
    card: list["CardField"]


class CardField(CamelModel):
    model_config = ConfigDict(str_strip_whitespace=True)

    key_name: str
    key_value: str
    key_type: Literal["str", "float", "date"]
    padding: bool
    default_value: str
    show_empty: bool
    prefix: Optional[str] = None
    postfix: Optional[str] = None

    @field_validator("key_value", mode="before")
    @classmethod
    def _key_value_validator(cls, v: str):
        assert v.startswith("properties.options.")
        return v.replace("properties.options.", "")

    @field_validator("key_type", mode="before")
    @classmethod
    def _key_type_validator(cls, v: str):
        alias_dict = {
            "text": "str",
            "number": "float",
            "date": "date",
            "href": "str",
        }
        return alias_dict[v]



import json
import re
from typing import Any, Generator, Optional, Type, Union, cast

import mercantile
import numpy as np
from httpx import Response
from shapely import MultiPolygon, Point, Polygon, to_geojson

from pynspd import asyncio_mock
from pynspd.client import get_client
from pynspd.errors import TooBigContour
from pynspd.schemas import Layer36048Feature, Layer36049Feature, NspdFeature
from pynspd.schemas.feature import Feat
from pynspd.schemas.responses import SearchResponse
from pynspd.types.enums import ThemeId


class Nspd:
    def __init__(self):
        self._client = get_client()

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self):
        """Окончание сессии"""
        self._client.close()

    @staticmethod
    def iter_cn(input_str: str) -> Generator[str, None, None]:
        """Извлечение кадастровых номеров из строки"""
        for cn in re.findall(r"\d+:\d+:\d+:\d+", input_str):
            yield cn

    @staticmethod
    def _cast_feature_to_layer_def(
        raw_feature: Optional[NspdFeature], layer_def: Type[Feat]
    ) -> Optional[Feat]:
        """Приведение базовой фичи к фиче конкретного слоя"""
        if raw_feature is None:
            return None
        feature = raw_feature.cast(layer_def)
        return feature

    @staticmethod
    def _cast_features_to_layer_defs(
        raw_features: Optional[list[NspdFeature]], layer_def: Type[Feat]
    ) -> Optional[list[Feat]]:
        """Аналог `_cast_feature_to_layer_def` для списка фичей"""
        if raw_features is None:
            return None
        features = [i.cast(layer_def) for i in raw_features]
        return features

    @staticmethod
    def _validate_response(response: Response) -> Optional[list[NspdFeature]]:
        response.raise_for_status()
        features = response.json()["features"]
        if len(features) == 0:
            return None
        return [NspdFeature.model_validate(i) for i in features]

    def _search(self, params: dict[str, Any]) -> Optional[SearchResponse]:
        r = self._client.get("/api/geoportal/v2/search/geoportal", params=params)
        if r.status_code == 404:
            return None
        r.raise_for_status()
        return SearchResponse.model_validate(r.json())

    def _search_one(self, params: dict[str, Any]) -> Optional[NspdFeature]:
        response = self._search(params)
        if response is None:
            return None
        features = response.data.features
        assert len(features) == 1
        return features[0]

    def search_by_theme(
        self, query: str, theme_id: ThemeId = ThemeId.REAL_ESTATE_OBJECTS
    ) -> Optional[NspdFeature]:
        """Глобальный поисковой запрос

        Args:
            query (str): поисковой запрос
            theme_id (int): вид объекта (кадастровое деление, объект недвижимости и т.д.)

        Returns:
            Optional[SearchResponse]: положительный ответ от сервиса, либо None, если ничего не найдено
        """
        return self._search_one(
            params={
                "query": query,
                "thematicSearchId": theme_id.value,
            }
        )

    def search_by_layers(self, query: str, *layer_ids: int) -> Optional[NspdFeature]:
        """Поисковой запрос по указанным слоям

        Args:
            query (str): поисковой запрос
            *layer_ids (int): id слоев, в которых будет производиться поиск

        Returns:
            Optional[SearchResponse]: положительный ответ от сервиса, либо None, если ничего не найдено
        """
        return self._search_one(
            params={
                "query": query,
                "layersId": layer_ids,
            }
        )

    def search_by_model(self, query: str, layer_def: Type[Feat]) -> Optional[Feat]:
        """Поиск одного объекта по определению слоя

        Определение слоя можно
        1) импортировать из `pynspd.schemas`, зная его id
        ```
        from pynspd.schemas import Layer36048Feature    # 36048 - id слоя, которое мы подсмотрели из запросов к wms на сайте
        feature = api.search_by_model("77:06:0004002:7207", Layer36048Feature)
        feature.properties.options.cad_num    # IDE знает тип и подсказывает возможные свойства
        ```

        2) воспользоваться методом `NspdFeature.by_title`
        ```
        from pynspd.schemas import NspdFeature
        feature = api.search_by_model("77:06:0004002:7207", NspdFeature.by_title("Земельные участки из ЕГРН"))    # IDE знает весь перечень слоев и подсказывает ввод
        feature.properties.options.cad_num    # свойство будет так же доступно, но IDE уже не знает о нем
        ```

        Args:
            query (str): поисковой запрос
            layer_def (Type[Feat]): Определение слоя

        Returns:
            Optional[Feat]: валидированная модель слоя, если найдено
        """
        feature = self.search_by_layers(query, layer_def.layer_meta.layer_id)
        return self._cast_feature_to_layer_def(feature, layer_def)

    def search_zu(self, cn: str) -> Optional[Layer36048Feature]:
        """Поиск ЗУ по кадастровому номеру"""
        layer_def = cast(
            Type[Layer36048Feature], NspdFeature.by_title("Земельные участки из ЕГРН")
        )
        return self.search_by_model(cn, layer_def)

    def search_many_zu(self, cns_string: str) -> list[Optional[Layer36048Feature]]:
        """Поиск всех ЗУ, содержащихся в строке"""
        cns = list(self.iter_cn(cns_string))
        features = asyncio_mock.gather(*[self.search_zu(cn) for cn in cns])
        return features

    def search_oks(self, cn: str) -> Optional[Layer36049Feature]:
        """Поиск ОКС по кадастровому номеру"""
        layer_def = cast(Type[Layer36049Feature], NspdFeature.by_title("Здания"))
        return self.search_by_model(cn, layer_def)

    def search_many_oks(self, cns_string: str) -> list[Optional[Layer36049Feature]]:
        """Поиск всех ОКС, содержащихся в строке"""
        cns = list(self.iter_cn(cns_string))
        features = asyncio_mock.gather(*[self.search_oks(cn) for cn in cns])
        return features

    def search_in_contour(
        self,
        countour: Union[Polygon, MultiPolygon],
        *category_ids: int,
        epsg: int = 4326,
    ) -> Optional[list[NspdFeature]]:
        """Поиск объектов в контуре по id категорий слоев

        Args:
            countour (Union[Polygon, MultiPolygon]): Геометрический объект с контуром
            category_ids (int): id категорий слоев
            epsg (int, optional): Система координат контура. Defaults to 4326.

        Returns:
            Optional[list[Feat]]: Список объектов, пересекающихся с контуром, если найден хоть один
        """
        feature_geojson = json.loads(to_geojson(countour))
        feature_geojson["crs"] = {
            "type": "name",
            "properties": {"name": f"EPSG:{epsg}"},
        }
        payload = {
            "categories": [{"id": id_} for id_ in category_ids],
            "geom": {
                "type": "FeatureCollection",
                "features": [
                    {"geometry": feature_geojson, "type": "Feature", "properties": {}}
                ],
            },
        }
        response = self._client.post(
            "/api/geoportal/v1/intersects",
            params={"typeIntersect": "fullObject"},
            json=payload,
        )
        if response.status_code == 500 and response.json()["code"] == 400004:
            raise TooBigContour
        return self._validate_response(response)

    def search_in_contour_by_model(
        self,
        countour: Union[Polygon, MultiPolygon],
        layer_def: Type[Feat],
        epsg: int = 4326,
    ) -> Optional[list[Feat]]:
        """Поиск объектов в контуре по определению слоя

        Args:
            countour (Union[Polygon, MultiPolygon]): Геометрический объект с контуром
            layer_def (Type[Feat]): Модель слоя
            epsg (int, optional): Система координат контура. Defaults to 4326.

        Returns:
            Optional[list[Feat]]: Список объектов, пересекающихся с контуром, если найден хоть один
        """
        raw_features = self.search_in_contour(
            countour, layer_def.layer_meta.category_id, epsg=epsg
        )
        return self._cast_features_to_layer_defs(raw_features, layer_def)

    def search_zu_in_contour(
        self, countour: Union[Polygon, MultiPolygon], epsg: int = 4326
    ) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ в контуре

        Args:
            countour (Union[Polygon, MultiPolygon]): Геометрический объект с контуром
            epsg (int, optional): Система координат контура. Defaults to 4326.

        Returns:
            Optional[list[Layer36048Feature]]: Список объектов, пересекающихся с контуром, если найден хоть один
        """
        return self.search_in_contour_by_model(countour, Layer36048Feature, epsg=epsg)

    def search_oks_in_contour(
        self, countour: Union[Polygon, MultiPolygon], epsg: int = 4326
    ) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС в контуре

        Args:
            countour (Union[Polygon, MultiPolygon]): Геометрический объект с контуром
            epsg (int, optional): Система координат контура. Defaults to 4326.

        Returns:
            Optional[list[Layer36048Feature]]: Список объектов, пересекающихся с контуром, если найден хоть один
        """
        return self.search_in_contour_by_model(countour, Layer36049Feature, epsg=epsg)

    def search_at_point(self, pt: Point, layer_id: int) -> Optional[list[NspdFeature]]:
        """Поиск объектов слоя в точке

        Args:
            pt (Point):
            layer_id (int):

        Returns:
            Optional[list[NspdFeature]]: Список объектов, если найдены
        """
        TILE_SIZE = 512
        tile = mercantile.tile(
            pt.x, pt.y, zoom=24
        )  # zoom=24 должно быть достаточно для самого точного совпадения
        tile_bounds = mercantile.bounds(tile)
        i = np.interp(pt.x, [tile_bounds.west, tile_bounds.east], [0, TILE_SIZE])
        j = np.interp(pt.y, [tile_bounds.south, tile_bounds.north], [0, TILE_SIZE])
        bbox = ",".join(
            map(str, mercantile.xy_bounds(tile))
        )  # bbox в 3857, см. комментарий про CRS
        params = {
            "REQUEST": "GetFeatureInfo",
            "SERVICE": "WMS",
            "VERSION": "1.3.0",
            "INFO_FORMAT": "application/json",
            "FORMAT": "image/png",
            "STYLES": "",
            "TRANSPARENT": "true",
            "QUERY_LAYERS": layer_id,
            "LAYERS": layer_id,
            "WIDTH": TILE_SIZE,
            "HEIGHT": TILE_SIZE,
            "I": int(i),
            "J": TILE_SIZE - int(j),  # отсчет координат для пикселей ведется сверху
            "CRS": "EPSG:3857",  # CRS для bbox
            # можно указать и 4326, но тогда и геометрия будет в 4326
            # Но в других методах мы всегда ждем 3857, поэтому оставляем
            "BBOX": bbox,
            "FEATURE_COUNT": "10",  # Если не указать - вернет только один, даже если попало на границу
        }
        response = self._client.get(f"/api/aeggis/v3/{layer_id}/wms", params=params)
        return self._validate_response(response)

    def search_at_point_by_model(
        self, pt: Point, layer_def: Type[Feat]
    ) -> Optional[list[Feat]]:
        """Поиск объектов слоя в точке (с типизацией)

        Args:
            pt (Point):
            layer_def (Type[Feat]): Тип слоя

        Returns:
            Optional[list[Feat]]: Типизированный список объектов, если найдены
        """
        raw_features = self.search_at_point(pt, layer_def.layer_meta.layer_id)
        return self._cast_features_to_layer_defs(raw_features, layer_def)

    def search_zu_at_point(self, pt: Point) -> Optional[list[Layer36048Feature]]:
        """Поиск ЗУ в точке"""
        return self.search_at_point_by_model(pt, Layer36048Feature)

    def search_oks_at_point(self, pt: Point) -> Optional[list[Layer36049Feature]]:
        """Поиск ОКС в точке"""
        return self.search_at_point_by_model(pt, Layer36049Feature)



from typing import ClassVar, Generator, Generic, Optional, Type, TypeVar, overload

from geojson_pydantic import Feature

from pynspd.errors import UnknownLayer
from pynspd.schemas.geometries import Geometry
from pynspd.schemas.layer_configs import LayerNode
from pynspd.schemas.properties import OptionProperties, Properties
from pynspd.types._autogen_layers import LayerTitle

Props = TypeVar("Props", bound="Properties")
Geom = TypeVar("Geom", bound="Geometry")
Feat = TypeVar("Feat", bound="_BaseFeature")


class _BaseFeature(Feature[Geom, Props], Generic[Geom, Props]):
    layer_meta: ClassVar[LayerNode]


class NspdFeature(_BaseFeature[Geometry, Properties[OptionProperties]]):
    """Базовая фича, приходящая из API, не привязанная к слою"""

    @classmethod
    def _iter_layer_defs(cls) -> Generator[Type["NspdFeature"], None, None]:
        root_class = cls.__base__.__base__
        for generic_subclass in root_class.__subclasses__():
            for subclass in generic_subclass.__subclasses__():
                meta = getattr(subclass, "layer_meta", None)
                if meta is not None:
                    yield subclass

    @classmethod
    def by_title(cls, title: LayerTitle) -> Type["NspdFeature"]:
        """Получение модели слоя по имени"""
        for layer_def in cls._iter_layer_defs():
            if layer_def.layer_meta.title == title:
                return layer_def
        raise UnknownLayer(title)

    @classmethod
    def by_category_id(cls, category_id: int) -> Type["NspdFeature"]:
        """Получение модели категории"""
        for layer_def in cls._iter_layer_defs():
            if layer_def.layer_meta.category_id == category_id:
                return layer_def
        raise UnknownLayer(category_id)

    @overload
    def cast(
        self, layer_def: None = None
    ) -> _BaseFeature[Geometry, Properties[OptionProperties]]: ...

    @overload
    def cast(self, layer_def: Type[Feat]) -> Feat: ...

    def cast(self, layer_def: Optional[Type[Feat]] = None):
        """Приведение объекта к одному из типов перечня определений слоев

        Args:
            layer_def (Optional[Type[Feat]], optional):
                Класс определение слоя. Если не указан, то попытается определить по свойствам. Defaults to None.

        Raises:
            UnknownLayer: Не удалось определить тип слоя по свойствам

        Returns:
            Feat: объект, приведенный к типу его слоя
        """
        if layer_def is None:
            assert self.properties is not None
            try:
                layer_def = self.by_title(self.properties.category_name)
            except UnknownLayer:
                # скрытый слой, пробуем определить свойства по категории
                similiar_def = self.by_category_id(self.properties.category)
                props_def = similiar_def.model_fields["properties"].annotation
                layer_def = _BaseFeature[Geometry, props_def]
        return layer_def.model_validate(self.model_dump(by_alias=True))



import ssl

import httpx

# Нет гарантий, что установлены сертификаты Минцифры, поэтому принудительно отключает проверку ssl
SSL_CONTEXT = ssl._create_unverified_context()
SSL_CONTEXT.set_ciphers("ALL:@SECLEVEL=1")


def get_async_client() -> httpx.AsyncClient:
    return httpx.AsyncClient(
        base_url="https://nspd.gov.ru",
        verify=SSL_CONTEXT,
        timeout=5,
        headers={
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:133.0) Gecko/20100101 Firefox/133.0",
            "Referer": "https://nspd.gov.ru",
        },
    )


def get_client() -> httpx.Client:
    return httpx.Client(
        base_url="https://nspd.gov.ru",
        verify=SSL_CONTEXT,
        timeout=5,
        headers={
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:133.0) Gecko/20100101 Firefox/133.0",
            "Referer": "https://nspd.gov.ru",
        },
    )



import re
from pathlib import Path

SUBS = [
    ("import asyncio", "from pynspd import asyncio_mock"),
    ("async def", "def"),
    ("async with", "with"),
    ("await ", ""),
    ("async for", "for"),
    ("get_async_client", "get_client"),
    ("__aenter__", "__enter__"),
    ("__aexit__", "__exit__"),
    ("AsyncClient", "Client"),
    ("AsyncNspd", "Nspd"),
    ("aclose", "close"),
    ("@pytest.mark.asyncio", ""),
    ("@pytest_asyncio", "@pytest"),
    ("asyncio", "asyncio_mock"),
]
COMPILED_SUBS = [
    (re.compile(r"(^|\b)" + regex + r"($|\b)"), repl) for regex, repl in SUBS
]


def unasync_file(file: Path):
    in_file = file.resolve()
    out_file = Path(str(in_file).replace("_async", "_sync"))
    out_file.parent.mkdir(parents=True, exist_ok=True)
    with (
        open(in_file, "r", encoding="utf-8") as in_stream,
        open(out_file, "w", encoding="utf-8") as out_stream,
    ):
        for line in in_stream:
            for pat, repl in COMPILED_SUBS:
                line = pat.sub(repl, line)
            out_stream.write(line)


def main():
    cwd = Path.cwd().resolve()
    folders = [cwd / "src", cwd / "tests"]
    for folder in folders:
        for async_folder in folder.glob("**/_async"):
            for file in async_folder.glob("**/*.py"):
                unasync_file(file)


if __name__ == "__main__":
    main()



import asyncio

from jinja2 import Template

from pynspd.client import get_async_client
from pynspd.schemas.layer_configs import Card, CardField, LayerNode, LayersTree

CLIENT = get_async_client()


TYPES_TEMPLATE = """# autogenerated
from typing import Literal


LayerTitle = Literal[
    {%- for layer in layers %}
    "{{ layer.title }}",
    {%- endfor %}
]
"""

SCHEMAS_TEMPLATE = """# autogenerated
from typing import Union, Annotated, Optional, override
from datetime import date

from geojson_pydantic import Feature
from pydantic import Field

from pynspd.schemas.geometries import Point, LineString, Polygon, MultiPolygon
from pynspd.schemas.layer_configs import LayerNode
from pynspd.schemas.feature import _BaseFeature
from pynspd.schemas.properties import Properties, OptionProperties


{% for category_id, fields in layers_fields.items() %}
class Options{{ category_id }}(OptionProperties): {% if fields|length == 0 %}...{% endif %}
    {%- for field in fields %}
    {{ field.key_value }}: Optional[{{field.key_type}}] = None
    \"\"\"{{field.key_name}}{% if field.postfix != None %} ({{ field.postfix }}){% endif %}\"\"\"
    {%- endfor %}
{% endfor %}


{% for layer in layers %}
class Layer{{ layer.layer_id }}Feature(_BaseFeature[
    {%- if layer.geometry_type == 'Polygon' %}
    Union[MultiPolygon, Polygon, Point],
    {% else %}
    {{ layer.geometry_type }},
    {% endif -%}
    Properties[Options{{ layer.category_id }}]
]):
    \"\"\"{{ layer.title }}\"\"\"
    layer_meta = LayerNode.model_validate({{ layer.model_dump(by_alias=True) }})
{% endfor %}
"""


def generate_files(layers: list[LayerNode], layers_fields: dict[int, Card]):
    output = Template(SCHEMAS_TEMPLATE).render(
        layers=layers, layers_fields=layers_fields
    )
    with open("src/pynspd/schemas/_autogen_features.py", "w", encoding="utf-8") as file:
        file.write(output)
    output = Template(TYPES_TEMPLATE).render(layers=layers)
    with open("src/pynspd/types/_autogen_layers.py", "w", encoding="utf-8") as file:
        file.write(output)


async def get_layer_tree() -> LayersTree:
    r = (
        await CLIENT.get("/api/geoportal/v1/layers-theme-tree", params={"themeId": 1})
    ).raise_for_status()
    tree = LayersTree.model_validate(r.json())
    return tree


async def get_category_card(category_id: int) -> tuple[int, list[CardField]]:
    r = (
        await CLIENT.get(f"/api/geoportal/v1/geom-card-display-settings/{category_id}")
    ).raise_for_status()
    card = Card.model_validate(r.json())
    fields = []
    for field in card.card:
        if field.key_value == "-":
            continue
        if field.key_value == "":
            continue
        if not field.key_value.isascii():
            continue
        if " " in field.key_value:
            continue
        fields.append(field)
    return category_id, fields


async def get_layers_fields(layers: list[LayerNode]) -> dict[int, list[CardField]]:
    layers_fields = {}
    categories = list(set([i.category_id for i in layers]))
    result = await asyncio.gather(*[get_category_card(i) for i in categories])
    layers_fields = dict(result)
    return layers_fields


async def main():
    tree = await get_layer_tree()
    layers = tree.layers
    layers_fields = await get_layers_fields(layers)
    generate_files(layers, layers_fields)


if __name__ == "__main__":
    asyncio.run(main())



from typing import Annotated, Generic, Union

import geojson_pydantic.geometries as pyd_geom
import shapely.geometry as shape_geom
from pydantic import Field

from pynspd.utils import BaseGeomT, from_3857_to_4326


class ShapeGeometry(Generic[BaseGeomT]):
    def to_shape(self, epsg4326: bool = True) -> BaseGeomT:
        """Конвертация в shapely-геометрию

        Args:
            epsg4326 (bool, optional): переводить ли в систему координат EPSG:4326. Defaults to True.

        Returns:
            BaseGeomT: геометрия
        """
        geom = shape_geom.shape(self)
        if epsg4326:
            geom = from_3857_to_4326(geom)
        return geom


class Point(pyd_geom.Point, ShapeGeometry[shape_geom.Point]): ...


class LineString(pyd_geom.LineString, ShapeGeometry[shape_geom.LineString]): ...


class Polygon(pyd_geom.Polygon, ShapeGeometry[shape_geom.Polygon]):
    def to_multi_shape(self, epsg4326: bool = True) -> shape_geom.MultiPolygon:
        return shape_geom.MultiPolygon([self.to_shape(epsg4326=epsg4326)])


class MultiPolygon(pyd_geom.MultiPolygon, ShapeGeometry[shape_geom.MultiPolygon]):
    def to_multi_shape(self, epsg4326: bool = True) -> shape_geom.MultiPolygon:
        return self.to_shape(epsg4326=epsg4326)


Geometry = Annotated[
    Union[
        Point,
        LineString,
        Polygon,
        MultiPolygon,
    ],
    Field(discriminator="type"),
]



from enum import Enum


class ThemeId(Enum):
    """Вид объекта (для поиска по категории)"""

    REAL_ESTATE_OBJECTS = 1
    """Объекты недвижимости"""

    CADASTRAL_DIVISION = 2
    """Кадастровое деление"""

    ADMINISTRATIVE_TERRITORIAL_DIVISION = 4
    """Административно-территориальное деление"""

    ZONES_AND_TERRITORIES = 5
    """Зоны и территории"""

    TERRITORIAL_ZONES = 7
    """Территориальные зоны"""

    COMPLEXES_OF_OBJECTS = 15
    """Комплексы объектов"""



# autogenerated
from typing import Literal

LayerTitle = Literal[
    "Кадастровые округа",
    "Территории объектов культурного наследия",
    "Кадастровые районы ",
    "Кадастровые кварталы",
    "Государственная граница Российской Федерации",
    "Субъекты Российской Федерации (линии)",
    "Субъекты Российской Федерации (полигоны)",
    "Муниципальные образования (полигональный)",
    "Муниципальные образования (линейный)",
    "Населённые пункты (полигоны)",
    "Населённые пункты (линии)",
    "Земельные участки из ЕГРН",
    "Земельные участки, образуемые по проекту межевания территории (линейный)",
    "Земельные участки, образуемые по проекту межевания территории (полигональный)",
    "Земельные участки, образуемые по схеме расположения земельного участка",
    "Земельные участки, выставленные на аукцион ",
    "Земельные участки, свободные от прав третьих лиц",
    "Земля для стройки ПКК",
    "Земля для туризма ПКК",
    "Здания",
    "Сооружения",
    "Объекты незавершенного строительства",
    "Единые недвижимые комплексы",
    "Предприятие как имущественный комплекс",
    "ЗОУИТ объектов культурного наследия",
    "ЗОУИТ объектов энергетики, связи, транспорта",
    "ЗОУИТ природных территорий",
    "ЗОУИТ охраняемых объектов и безопасности",
    "Иные ЗОУИТ",
    "Территориальные зоны",
    "Красные линии ",
    "Особо охраняемые природные территории ",
    "Охотничьи угодья ",
    "Лесничества",
    "Береговые линии (границы водных объектов) (полигональный)",
    "Береговые линии (границы водных объектов)(линейный)",
    "Особые экономические зоны",
    "Территории опережающего развития",
    "Зоны территориального развития в Российской Федерации",
    "Игорные зоны",
    "Территории выполнения комплексных кадастровых работ",
    "Территория проведения мероприятий по ликвидации накопленного вреда окружающей среде, образовавшегося в результате производства химической продукции в г. Усолье-Сибирское Иркутской области",
    "Негативные процессы",
    "Объекты туристского интереса",
    "Кадастровая стоимость объекта",
    "Удельный показатель кадастровой стоимости",
]



from typing import TypeVar

import pyproj
from shapely.geometry.base import BaseGeometry
from shapely.ops import transform

BaseGeomT = TypeVar("T", bound=BaseGeometry)

epsg4326_proj = pyproj.CRS("EPSG:4326")
epsg3857_proj = pyproj.CRS("EPSG:3857")

project_3857_to_4326 = pyproj.Transformer.from_crs(
    epsg3857_proj, epsg4326_proj, always_xy=True
).transform

project_4326_to_3857 = pyproj.Transformer.from_crs(
    epsg4326_proj, epsg3857_proj, always_xy=True
).transform


def from_3857_to_4326(geom: BaseGeomT) -> BaseGeomT:
    """Перепроецирование геометрии из EPSG:3857 в EPSG:4326"""
    return transform(project_3857_to_4326, geom)


def from_4326_to_3857(geom: BaseGeomT) -> BaseGeomT:
    """Перепроецирование геометрии из EPSG:4326 в EPSG:3857"""
    return transform(project_4326_to_3857, geom)
